function totalCutEnergy = ComputeEnergy_ClothingFaceSpeechUnique(idEnergy)
% Computes the id based energy of a system dealing with tracks present
% within a cut
global ENERGY
% i - across scenes; j - across shots; k - across tracks
numIDs = size(idEnergy,2);

% FACE generates face based energy for each track
% EF(i,j,k) = \mu(i,j,k)*f(i,j,k)' - a vector of values for each track
combinedFace = -sum(idEnergy.*ENERGY.faceEnergy,2);

% CLTH generates clothing based energy for each track
% EC(i,j,k) = \mu(i,j,k)*c(i,j,k)' - a vector of values for each track
combinedClth = -sum(idEnergy.*ENERGY.clthEnergy,2);

% FACE + CLTH combine the two IDs, with weights
% E(i,j) = wClth * \sum_k{EC(i,j,k)} + wFace * \sum_k{EF(i,j,k)}
IDs = ENERGY.faceWeight*combinedFace + ...
      ENERGY.clthWeight*combinedClth; 

% SPKR some positive energy to ensure that the speaker is included if possible
% (ones(vec) - sigmoid(\sum_k(\mu_k)))*\sigma'
idsInCut = sum(idEnergy,1);
penalizeSpeaker = (ones(1,numIDs) - mySigmoid(idsInCut,ENERGY.penalize_B))*ENERGY.spkrEnergy';
  
% SELF self energy of idEnergy - needs to be taken into account to restrict the
% combinations of ids, else all 1s will be always best situation
ENERGY.idSelfEnergy = ENERGY.selfWeight*sum(sum(idEnergy.^2));
% selfEnergy = ENERGY.selfWeight*sum(sum(idEnergy));

% totalEnergy = energy of IDs + self energy
ENERGY.totalCutEnergy = sum(IDs) + ENERGY.idSelfEnergy;
ENERGY.totalCutEnergy = ENERGY.totalCutEnergy + ENERGY.spkrWeight*penalizeSpeaker;

% which tracks occur together (same frames), they should not get the same id (except the case they are unknowns!)
uniquePenalty = 0;
for k=1:size(ENERGY.uniqueClthtracks)
    uniquePenalty = uniquePenalty + sum(idEnergy(ENERGY.uniqueClthtracks(k,3),1:5).*idEnergy(ENERGY.uniqueClthtracks(k,4),1:5));
end
ENERGY.totalCutEnergy = ENERGY.totalCutEnergy + ENERGY.uniqWeight*uniquePenalty;

% return this as an optimization objective value
totalCutEnergy = ENERGY.totalCutEnergy;


end