% code to learn the mean histogram datapoints of clothing for clusters with
% a quite sure ID assigned to it by the face clustering algorithm..
% this is also followed by a re-clustering of the clusters which are not so
% sure right now...

% clear;
% clc;
% close all;

function clusterAssignReassign_nBestResults_paper(epinum, allWriteNumbers, idThresh)
%% Basic handling

% epinum = '1'; 
% allWriteNumbers = {'1','2','3','4','5'}; % };%

tic;

% load the precomputed histograms...
load(['/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClothingFeatures/bodyImFeats_E0',epinum,'.mat']);

% load FACES for associating the results
% load(['/home/mtapaswi/MS_Thesis/FinalDATA/FACES/E',epinum,'.mat']);

% load the all info storage file, can be used later with dlmread for getting results easily
% structure: epinum, writenum, idThresh, assignedCorrect, totalAssigned, unassignedCorrect, totalUnassigned, finalPercentage
fidFinal = fopen('/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClusteredClothingIndividual/clusterAssignReassign_nBestResults_paper/AllResults.txt','a');

%% Parameters

% 1='Leonard'; 2='Sheldon'; 3='Penny'; 4='Raj'; 5='Howard'; 6='UNKNOWN'; 7='FalsePositive'; 8='NO_FACES_FOUND'; 9=''; 
% 6 -> unknown id detected; 7 -> false positive face detection; 8 -> face classifier svm gave nothing; 9 -> no face found in body image
possibleResultIDs = {'Leonard', 'Sheldon', 'Penny', 'Raj', 'Howard', 'UNKNOWN', 'FalsePositive', 'NO_FACES_FOUND', ''}; 

idThresh = 0.6; % this ratio of cluster faces (1:6) should belong to chosen ID
totalThresh = 0.10; % this ratio of all possible IDs (including blank) should belong to chosen ID
stddevThresh = 0.6; % the histogram of chosen cluster's images should not exceed this

featlength = 64; % 4^3 3D histogram length...

%% Put all filenames in one cell array, easier to compare later with strcmp

% load the face and body framewise information
imDbPath = ['/home/mtapaswi/Database/bodyImages/E0',epinum,'/'];
allIms = dir([imDbPath,'*.png']);

fprintf('Reading all filenames...\n');
allNames = cell(length(allIms),1);
for k=1:length(allIms)
    allNames{k} = allIms(k).name;
end

%% Start processing for each write-number 
for writeIdx=1:length(allWriteNumbers)

    writeNumber = allWriteNumbers{writeIdx};
    epifolder = ['/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClusteredClothingIndividual/E',epinum,'_',writeNumber,'/'];

	% load the clustered and assigned results
    load(['/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClusteredClothingIndividual/FaceCorrespondence_paper/E',epinum,'_',writeNumber,'.mat']);    

    % open a new file to write results in
    fid = fopen(['/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClusteredClothingIndividual/clusterAssignReassign_nBestResults_paper/E',epinum,'_',writeNumber,'_idThresh',num2str(idThresh*100),'_totalThresh',num2str(totalThresh*100),'.txt'],'w');
    fprintf(fid, '%s\n',datestr(now));

    fprintf(fid, '---------------------------------------------------------------------------------------------------------\n');
    fprintf(fid, 'Episode %s | Write Number %s\n',epinum,writeNumber);
    fprintf('Episode %s | Write Number %s\n',epinum,writeNumber);
    
   
    %% See which cluster contains whose face and how many... 
    % basically convert the text results (from loaded data in FaceCorrespondence) to number based results
    
    totalScenes = size(clusteredImageFaceResult,1);
    maxClusters = size(clusteredImageFaceResult,2);    
    
    fprintf(fid, 'Getting count of result ID across clusters...\n');
    
    idCounts = zeros(totalScenes,maxClusters,length(possibleResultIDs));
    
    for i=1:totalScenes % num scenes
        for j=1:maxClusters % max num clusters in the scene
            
            % just ensure there are files in that cluster
            if ~isempty(clusteredImageFaceResult(i,j).fileName)
%                 fprintf('*** SCENE %d :: CLUSTER %d ***\n',i,j);
                for k=1:length(clusteredImageFaceResult(i,j).ResultID)
                    whoIsIt = strcmpi(clusteredImageFaceResult(i,j).ResultID{k}, possibleResultIDs);
                    idCounts(i,j,whoIsIt==1) = idCounts(i,j,whoIsIt==1) + 1;
                end
            end
        end
    end
       
    %% Compute Cluster histogram (feature) statistics - standard deviation and mean
    % standard deviation another feature that could be used for checking whether to assign the cluster with an ID
    fprintf(fid, 'Compute Cluster histogram features standard deviation...\n');
    
    clusterStddev = zeros(totalScenes,maxClusters,featlength);
    clusterMean = zeros(totalScenes,maxClusters,featlength);
    
    for i=1:totalScenes % num scenes
        for j=1:maxClusters % max num clusters in the scene
            if sum(squeeze(idCounts(i,j,:)))==0; continue; end % it            
            
            fprintf(fid, '|%d:%02d|',i,j);
            
            % get the images in this folder, compare with database folder
            % get the indices corresponding to the cluster image's histograms
            clusterFolder = [epifolder, sprintf('Scene%d/Cluster%02d/',i,j)];
            clusterIms = dir([clusterFolder, '*.png']);

            clusterHist = zeros(length(clusterIms),featlength);
            for k=1:length(clusterIms)
                filenumber = strcmpi(allNames, clusterIms(k).name);
                clusterHist(k,:) = hist3Dim(filenumber,:);
            end

            clusterStddev(i,j,:) = sqrt(var(clusterHist));
            clusterMean(i,j,:) = mean(clusterHist);
%             fprintf('Numimages = %3d | Standard deviation = %.4f\n',length(allIms),histStd(i,j));
        end
        fprintf(fid, '\n');
    end
    
    %% Make a decision whether a cluster is good to learn or not good enough
    fprintf(fid, 'Making decisions whether cluster IDs are good...\n');
        
    assignedClusters.is = NaN*ones(totalScenes, maxClusters);    
    assignedClusters.NumberID = zeros(totalScenes, maxClusters);
    
    for i=1:totalScenes % num scenes
        for j=1:maxClusters % max num clusters in the scene
            useVec = squeeze(idCounts(i,j,:));
            totalIms = sum(useVec); 
            if totalIms==0, continue; end % it means that cluster is empty            
            totalIDlabels = sum(useVec(1:6));
            idVec = useVec(1:6)/totalIDlabels;
            normalizedVec = useVec/totalIms;
            
            satisfyIdThresh = intersect(find(idVec>=idThresh),find(normalizedVec>=totalThresh));
            if length(satisfyIdThresh)>1, satisfyIdThresh = []; end % if it finds more than 1 id satisfying conditions, cluster is bad. remove all...
            
            fprintf(fid, '\tSCENE %d :: CLUSTER %2d : ',i,j);
%             disp([useVec' NaN sum(useVec(1:6)) sum(useVec)]);
            if ~isempty(satisfyIdThresh) && sum(squeeze(clusterStddev(i,j,:))) < stddevThresh
                fprintf(fid, '*** Assign to cluster = %s\n',possibleResultIDs{satisfyIdThresh});
                assignedClusters.NameID{i,j} = possibleResultIDs{satisfyIdThresh};
                assignedClusters.NumberID(i,j) = satisfyIdThresh;
                assignedClusters.is(i,j) = 1;
            else
                if sum(squeeze(clusterStddev(i,j,:))) > stddevThresh, fprintf(fid, 'stddevThresh constrained. '); end
                if any(normalizedVec>=totalThresh), fprintf(fid, 'totalThresh constrained. '); end
                if any(idVec>=idThresh), fprintf(fid, 'idThresh constrained. '); end
                assignedClusters.is(i,j) = 0;
                fprintf(fid, ' Cannot assign.\n');
            end
            
            % if cluster is assigned an ID, then need to calculate probs, else it will be taken care of later in the unassigned clusters section
            if assignedClusters.is(i,j)
                % assign the probability of being an ID to every image of the cluster to the CLOTHING thingy
                clusterFolder = [epifolder, sprintf('Scene%d/Cluster%02d/',i,j)];
                clusterIms = dir([clusterFolder, '*.png']);

                % note that idVec forms a probability measure directly
                % because it contains the number of faces corresponding to
                % each person, associated to the whole cluster. 
%                 disp(idVec')

                for eachIm=1:length(clusterIms)
                    % pick the filenumber corresponding to the filename
                    filenumber = strcmpi(allNames, clusterIms(eachIm).name);
                    CLOTHING(filenumber).filename = clusterIms(eachIm).name;
                    CLOTHING(filenumber).Probability = idVec';
                end
            end            
        end
    end
    
    totalClusters = sum(sum(~isnan(assignedClusters.is)));
    totalAssigned = sum(sum(assignedClusters.is(~isnan(assignedClusters.is))));
    totalUnassigned = totalClusters - totalAssigned;
    
    fprintf(fid, 'Assigned %.3f = %d / %d clusters an ID\n',totalAssigned/totalClusters,totalAssigned,totalClusters);
    
    %% Pick the cluster mean data point for assigned clusters
    % for unassigned clusters, go through each image and find to which mean point does it have minimum distance    
    % cluster means were already computed when computing standard deviation for checking goodness
    % also check how good are the cluster assignments...

    assignedCorrect = 0; assignedWrong = 0; unassignedCorrect = 0; unassignedWrong = 0;
    
    fprintf(fid, 'Selecting Cluster feature mean data points and checking goodness of cluster label...\n');
    assignedClusterMeans = zeros(totalAssigned, featlength+2);
    
    assignedIdx = find(assignedClusters.is==1);
    [sceneNum, clusterNum] = ind2sub([size(idCounts,1),size(idCounts,2)], assignedIdx);
    
    for k=1:totalAssigned
        i = sceneNum(k); j = clusterNum(k);
        fprintf(fid, '\t%3d. SCENE %d :: CLUSTER %2d :: ID %10s\n',k,i,j,assignedClusters.NameID{i,j});
        
        % store this thing's statistic as a data point for other comparisons in the scene (across scenes?)
        assignedClusterMeans(k,1) = i;
        assignedClusterMeans(k,2) = j;
        assignedClusterMeans(k,3:end) = clusterMean(i,j,:);
        
        % check how many are correctly assigned
        clusterFolder = [epifolder, sprintf('Scene%d/Cluster%02d/',i,j)];
        clusterIms = dir([clusterFolder, '*.png']);
        
%         for eachIm=1:length(clusterIms)
            % pick the histogram corresponding to the filename
%             filenumber = strcmpi(allNames, clusterIms(eachIm).name);
%             if strcmpi(clusteredImageFaceResult(i,j).groundTruth{eachIm}, assignedClusters.NameID{i,j})
%                 assignedCorrect = assignedCorrect + 1;
%             else
%                 assignedWrong = assignedWrong + 1;
%             end            
%         end
    end
    assignedNumpics = assignedCorrect + assignedWrong;
    
    %% Compare non-assigned clusters with computed cluster centers
    % also try and include the face information for images that have it
    % but basically we say that they DO NOT form a cluster, its a mixture
    
    fprintf(fid, 'Comparing the rest of the ID unassigned clusters...\n');
    
    unassignedIdx = find(assignedClusters.is==0);
    [sceneNum, clusterNum] = ind2sub([size(idCounts,1),size(idCounts,2)], unassignedIdx);
    
    % if there are no unassigned clusters, then just put to blank
    if ~totalUnassigned, unassignedCluster = []; end
    
    for k=1:totalUnassigned
        i = sceneNum(k); j = clusterNum(k);

        fprintf(fid, 'SCENE %d :: CLUSTER %2d\n',i,j);
        % get the images in this folder, compare with database folder
        % get the indices corresponding to the cluster image's histograms
        clusterFolder = [epifolder, sprintf('Scene%d/Cluster%02d/',i,j)];
        clusterIms = dir([clusterFolder, '*.png']);
        
        % get the row number corresponding to the i'th scene in the histogram stores
        % get the row number corresponding to the i'th scene in the histogram stores
        sceneAssignedTo = find(assignedClusterMeans(:,1)==i); % uses only i-th scene
        if isempty(sceneAssignedTo)
            % uses (i)th (i+1)th and (i-1)th scenes too
            sceneAssignedTo = union(find(assignedClusterMeans(:,1)==i),union(find(assignedClusterMeans(:,1)==i),find(assignedClusterMeans(:,1)==i+1))); 
            fprintf(fid, 'WARNING: Using clothing information also from adjacent scenes\n');
            if isempty(sceneAssignedTo)
                % uses all scenes
                sceneAssignedTo = 1:size(assignedClusterMeans,1); 
                fprintf(fid, 'WARNING: Using clothing information from all scenes\n');
            end
        end
        
        
        for eachIm=1:length(clusterIms)
            % pick the histogram corresponding to the filename
            filenumber = strcmpi(allNames, clusterIms(eachIm).name);
            pickHist = [i, j, hist3Dim(filenumber, :)];
            
            % once picked, compute distances with each of the assignedClusterMeans vectors and find out who's closest
            dist2AllMeans = pdist2(pickHist(3:end), assignedClusterMeans(sceneAssignedTo,3:end));
            [sortDist,sortIdx] = sort(dist2AllMeans);
            
%             probMeasure = exp(-sortDist);
            probMeasure = exp(-1*cumsum(sortDist)./sortDist); % convert those distances to some sort of probability
            probMeasure = probMeasure/sum(probMeasure);
            
            assignedToScene = zeros(1,length(probMeasure));
            assignedToCluster = zeros(1,length(probMeasure));
            NameIDs = cell(1,length(probMeasure));
            
            % evaluate the ID and the associated probability for all
%             for eachId = 1:sum(probMeasure>0.1) % basically puts a cutoff at the number of assignments to make
            for eachId = 1:length(probMeasure) % uses distance with all clusters...
                assignedToScene(eachId) = assignedClusterMeans(sceneAssignedTo(sortIdx(eachId)),1);
                assignedToCluster(eachId) = assignedClusterMeans(sceneAssignedTo(sortIdx(eachId)),2);
                NameIDs{eachId} = assignedClusters.NameID{assignedToScene(eachId), assignedToCluster(eachId)};
            end
            
            % combine same names, and add probability measures
            allProbs = zeros(1,6);
            for eachId = 1:length(NameIDs)
                whoIsIt = strcmpi(NameIDs{eachId}, possibleResultIDs);
                allProbs(whoIsIt==1) = allProbs(whoIsIt==1) + probMeasure(eachId);
            end
    
            % Probability-based results assignment for each file in the unassigned clusters
            CLOTHING(filenumber).filename = clusterIms(eachIm).name;
            CLOTHING(filenumber).Probability = allProbs;

            % Check who has max prob to see "accuracy", although feeding to
            % model will use the probabilities finally...
            [~,bestId] = max(allProbs);
%             if strcmpi(clusteredImageFaceResult(i,j).groundTruth{eachIm}, possibleResultIDs{bestId})
%                 unassignedCorrect = unassignedCorrect + 1;
%             else
%                 unassignedWrong = unassignedWrong + 1;
%             end            
        end
    end
    
    unassignedNumpics = unassignedCorrect + unassignedWrong;
    
    %% ID and Unknown differentiation
    
    IDwrong = 0; IDcorrect = 0; UnknownWrong = 0; UnknownCorrect = 0;
    
    for k=1:length(CLOTHING)
        if isempty(CLOTHING(k).Probability), continue; end
        
        GroundTruthID = CLOTHING(k).filename(1:strfind(CLOTHING(k).filename,'_')-1);
        [~,bestId] = max(CLOTHING(k).Probability);

        % check if the groundtruth is unknown and then classify
        if strcmpi(GroundTruthID,'Unknown')
            if strcmpi(GroundTruthID, possibleResultIDs{bestId}), UnknownCorrect = UnknownCorrect + 1;
            else UnknownWrong = UnknownWrong + 1; end
        
        % means groundtruth is one of the ids, and now, check which is correct   
        else
            if strcmpi(GroundTruthID, possibleResultIDs{bestId}), IDcorrect = IDcorrect + 1;
            else IDwrong = IDwrong + 1; end
        end
    end
    
    %% FINAL FINAL FINAL :)
    totalCorrect = assignedCorrect + unassignedCorrect;
    totalNumpics = assignedNumpics + unassignedNumpics;
    
    fprintf(fid, 'ASSIGNED CORRECT CLASSIFICATION COUNT: %d / %d || Percentage: %.3f\n',assignedCorrect,assignedNumpics,100*assignedCorrect/assignedNumpics);
    fprintf(fid, 'UNASSIGNED CORRECT CLASSIFICATION COUNT: %d / %d || Percentage: %.3f\n',unassignedCorrect,unassignedNumpics,100*unassignedCorrect/unassignedNumpics);
    
    fprintf(fid, 'FINAL CORRECT CLASSIFICATION COUNT: %d / %d || Percentage: %.3f\n',totalCorrect, totalNumpics, 100*totalCorrect/totalNumpics);    
    
    save(['/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClusteredClothingIndividual/clusterAssignReassign_nBestResults_paper/E',epinum,'_',writeNumber,'_idThresh',num2str(idThresh*100),'_totalThresh',num2str(totalThresh*100),'.mat'],'CLOTHING','possibleResultIDs');
    
    % epinum, writenum, idThresh, assignedCorrect, totalAssigned, unassignedCorrect, totalUnassigned, finalPercentage, IDcorrect, IDwrong, UnknownCorrect, UnknownWrong
    fprintf(fidFinal, '%s, %s, %2d, %4d, %4d, %4d, %4d, %5.3f, %4d, %4d, %4d, %4d\n',epinum,writeNumber,100*idThresh,assignedCorrect,assignedNumpics,unassignedCorrect,...
                                                                                     unassignedNumpics,100*totalCorrect/totalNumpics,IDcorrect,IDwrong,UnknownCorrect,UnknownWrong);

    %% CLOTHING check to see all files have info on them...
    fprintf(fid, 'Now writing about files which have no info, but have a histogram!\n');
    for k=1:length(CLOTHING)
        if isempty(CLOTHING(k).Probability)
            % implies this image should be removed from clustering because
            % too little part was inside the frame...
            
            % check if its histogram is empty, else something is wrong...
            if sum(hist3Dim(k,:))
                fprintf(fid, '\t%4d is empty\n',k);
                keyboard;
            end
        end
    end
    fclose(fid);
    
end

fclose(fidFinal);
fprintf('Elapsed time is %.3f\n',toc);
