% Energy based optimization on the model
% Working right now in isolated cuts
% This adds the face track results to the base of clothing track results and also
% includes the penalty for speaker not being included

% There is a permanent mismatch between probabilities of clothing and face
% Face IDs ordered - Sheldon Leonard Penny Howard Raj UNKNOWN
% Clothing ordered - Leonard Sheldon Penny Raj Howard UNKNOWN
% Fixed using flipClothingProbs()

% clear;
% clc;
% close all;

%% Loading Model Info
function energyOptCuts_faceInClothingSpeechUnique(epinum)   
% epinum = '1';

IDs = {'Sheldon', 'Leonard', 'Penny', 'Howard', 'Raj', 'Unknown'};
nStates = length(IDs); % number of components for each vector - 6

% Load Clothing tracks information
load(['../../FinalDATA/CLOTHINGTRACKS_paper/E',epinum,'.mat']); 

% Load Face information
load(['../../FinalDATA/FACES_normalized/E',epinum,'.mat']); 

% Load Speakers information
load(['../../FinalDATA/SPEAKERS/E',epinum,'.mat']); 

% Load the faces and clothing correspondences
load(['../../FinalDATA/FaceClothingCorrespondence/ClothingTrack/E',epinum,'.mat']);

%% Model Initialisation stuff

tinyValue = 0.05; % to allow others to exist..
SPKR_timeB = 0.3; % mid-index for speaker time score sigmoid fn.
% SPKR_pntgB = 0; % mid-index for speaker percentage score sigmoid fn.
SPKR_penalize_B = 0; % mid-index for the speaker penalization sigmoid (inside the ComputeEnergy functions)

FACE_multiplier = 1; % faceID weight
CLTH_multiplier = 1; % clothingID weight
SPKR_multiplier = 2; % speakerID weight
SELF_multiplier = 1; % self energy weight
UNIQ_multiplier = 2; % unique ID weight

threshFalsePositive = 0; % if none are above 0.4 declare as false positive

global ENERGY

%% Iterate over all cuts, and each image in cut

totalError = 0; 
totalFalsePositives = 0; FPcounter = 0;
totalClthError = 0; ClthErrorInc = 0;
allError = 0;
totalIms = 0;

evalCuts = length(faceTracksInShot);
setNum = 1;

for cutNum = 1:evalCuts
    fprintf('-------------------- CUT NUMBER: %3d --------------------\n',cutNum);
    
    % Clothing tracks in the shot
    clthTracks = clothingTracksInShot{cutNum};
    numTracks = length(clthTracks);
    cutGroundTruth = zeros(1,numTracks);
    
    if numTracks==0
        fprintf('NO TRACKS FOUND\n');
        continue;
    end
    
    % Initialization
    faceEnergy = tinyValue*ones(numTracks,nStates);
    clthEnergy = tinyValue*ones(numTracks,nStates);
    spkrEnergy = tinyValue*ones(1,nStates);    
    idEnergy = tinyValue*ones(numTracks,nStates);
    
    % Bounds for optimization
    lowerBound = zeros(size(idEnergy));
    upperBound = 1e10*ones(size(idEnergy));
           
   %% Get "energy" value from scores

    for trackNum = 1:numTracks
        chosenTrack = clthTracks(trackNum);
            
        % If track exists, get an influence for it
        if CLOTHINGTRACKS(chosenTrack).exist==1
            ModUniqResults(chosenTrack).exist = 1;
            
            % Clothing image stuff
            outProbs = flipClothingProbs(CLOTHINGTRACKS(chosenTrack).FinalProbs);
            clthEnergy(trackNum,:) = clthEnergy(trackNum,:) + outProbs;
            
            % direct clothing result errors
            [~,idx] = max(outProbs);
            if ~strcmpi(IDs{idx},CLOTHINGTRACKS(chosenTrack).groundTruth)
                totalClthError = totalClthError + 1;
            end
            ClthErrorInc = ClthErrorInc + 1;
            
            fprintf('\tCut %3d, Track %4d: %10s ',cutNum,chosenTrack,CLOTHINGTRACKS(chosenTrack).groundTruth);
            
            cutGroundTruth(trackNum) = find(strcmpi(CLOTHINGTRACKS(chosenTrack).groundTruth,IDs)==1);
            
        else
            ModUniqResults(chosenTrack).exist = 0;
            cutGroundTruth(trackNum) = 7; % False Positive indicator
            fprintf('\tCut %3d, Track %4d: FALSE POSITIVE ',cutNum,chosenTrack);
            continue;
        end
                    
        % Face track stuff
        % check whether the clothing image contains any face tracks
        faceTracks = FaceClothingTrackCorrespondence(chosenTrack).faceTrackNum;
        for eachFaceTrack = 1:length(faceTracks)
            if ~isempty(faceTracks(eachFaceTrack))
            % then we have a face track corresponding to clothing image
                for i=1:length(FACES(faceTracks(eachFaceTrack)).nbestID)
                    whoseID = strcmpi(FACES(faceTracks(eachFaceTrack)).nbestID{i},IDs);
                    if ~any(whoseID), fprintf('NotEval'); end
                    scoreToEnergy = FACES(faceTracks(eachFaceTrack)).nbestScore(i);
                    faceEnergy(trackNum,whoseID) = faceEnergy(trackNum,whoseID) + scoreToEnergy;
                end
                fprintf('\t Face track exists.');
            end
        end
        fprintf('\n');
    end
    
    % Speaker stuff
    if cutNum<=length(SPEAKERS)
        numSpeakers = length(SPEAKERS(cutNum).ID);
        for speakerNum = 1:numSpeakers
            if strcmp(SpeakerNames{SPEAKERS(cutNum).ID(speakerNum)},'Other')
                % SpeakerID's "Other" = FaceID's "UNKNOWN"
                whoseID = logical([0 0 0 0 0 1]);
            else
                whoseID = strcmp(SpeakerNames{SPEAKERS(cutNum).ID(speakerNum)},IDs);
            end
            timeToEnergy = mySigmoid(SPEAKERS(cutNum).time(speakerNum), SPKR_timeB);
            pntgToEnergy = SPEAKERS(cutNum).pntg(speakerNum)/200;
            spkrEnergy(whoseID) = spkrEnergy(whoseID) + timeToEnergy + pntgToEnergy;
        end
    end
    
    % send numbers of which tracks are unique
    if cutNum > length(clothingTrackSetInfo)
        ENERGY.uniqueClthtracks = [];
    else
        ENERGY.uniqueClthtracks = clothingTrackSetInfo{cutNum};
    end

    
%     % check which tracks occur separately in this shot (for uniqueness)
%     if isempty(setxor(clthTracks, clothingTrackSetInfo{setNum}))
%         % means that all tracks in this shot are in the same set
%         ENERGY.uniqueBreakpt = length(clthTracks);
%         setNum = setNum + 1;
%     else
%         % its broken into 2 parts
%         inc = 1;
%         while intersect(clthTracks, clothingTrackSetInfo{setNum})
%             ENERGY.uniqueBreakpt(inc) = length(clothingTrackSetInfo{setNum});
%             setNum = setNum + 1; inc = inc + 1;
%         end
%     end
    
    %% Optimization Part
    % populate the global ENERGY structure used to compute and optimize

    ENERGY.faceEnergy = faceEnergy;
    ENERGY.clthEnergy = clthEnergy;
    ENERGY.spkrEnergy = spkrEnergy;
    ENERGY.penalize_B = SPKR_penalize_B;
    
    ENERGY.faceWeight = FACE_multiplier;
    ENERGY.clthWeight = CLTH_multiplier;
    ENERGY.spkrWeight = SPKR_multiplier;
    ENERGY.selfWeight = SELF_multiplier;
    ENERGY.uniqWeight = UNIQ_multiplier;
    ENERGY.initEnergy = ComputeEnergy_ClothingFaceSpeechUnique(idEnergy);
    
    % setting the options up
    options = optimset('Algorithm','active-set','Display','off');
    [optimIDs,fval,exitflag] = fmincon(@ComputeEnergy_ClothingFaceSpeechUnique,idEnergy,[],[],[],[],lowerBound,[],[],options);

% Creating a GlobalSearch and MultiStart object
%     problem = createOptimProblem('fmincon','x0',idEnergy,'objective',@compIdEnergy,'lb',lowerBound,'options',options);
%     gs = GlobalSearch; % GlobalSearch object
%     ms = MultiStart; ms.UseParallel = 'always'; % MultiStart object with parallel programming
%     [xminMS, fminMS, flagMS, outputMS, allminsMS] = run(ms,problem,10); % run MultiStart with 10 starting points...
%     [xminGS, fminGS, flagGS, outputGS, allminsGS] = run(ms,problem,10); % run GlobalSearch...
    
    ENERGY.optimIDs = optimIDs;
    ENERGY.finalEnergy = fval;
    
    
    for eachtrack = 1:length(clthTracks)
        ModUniqResults(clthTracks(eachtrack)).optimEnergy = optimIDs(eachtrack,:);
    end

        
    %% Cut Results
    disp('Face'); disp(faceEnergy);
    disp('Clothing'); disp(clthEnergy);
    disp('Speaker'); disp(spkrEnergy);
    disp('Optimal'); disp(optimIDs); 
    fprintf('From %.3f to %.3f\n',ENERGY.initEnergy,fval);
    
    whoIs = zeros(1,numTracks);
    for i=1:numTracks
        [assignID,loc] = max(optimIDs(i,:));
        
        if assignID<threshFalsePositive
            % if all IDs get a 0 in a track (then possibly false positive)
            whoIs(i) = 7;
            fprintf('%2d. ID: NO FACES FOUND, NO CLOTHING FOUND\n',i);
        else
            % find the maxima for each track and assign label to that ID    
            whoIs(i) = loc;
            
            % just a error indicator to find them easily! :)
            if whoIs(i)~=cutGroundTruth(i), fprintf('--> --> --> '); end
            fprintf('%2d. ID: %s\n',i,IDs{whoIs(i)});
        end
    end
    
    % computes the errors for the tracks which are not false positives
    nonFalseIDs = find(cutGroundTruth~=7);
    totalError =  totalError + length(nonzeros(whoIs(nonFalseIDs)-cutGroundTruth(nonFalseIDs)));
    
    % includes false positive classification errors
    allError = allError + length(nonzeros(whoIs - cutGroundTruth));
    totalFalsePositives = totalFalsePositives + (length(cutGroundTruth)-length(nonFalseIDs));
    
    fprintf('Clothing Errors: %3d | ID Errors: %3d | All Errors: %3d | FalsePositives: %3d\n',totalClthError,totalError,allError,totalFalsePositives);

    % Store everything in one big matrix!
    AllShotsEnergy(cutNum) = ENERGY;    
%     keyboard;

end

%% FINAL RESULT

% save(['../../FinalDATA/Results_CLOTHINGTRACKS_dt2/E',epinum,'.mat'],'AllShotsEnergy');

% save for prec-recall curves
save(['../../FinalDATA/Results_Paper/ModUniqE',epinum,'.mat'],'ModUniqResults');



fprintf('====================================================================\n');
fprintf('Episode number = %s\n',epinum);
fprintf('====================================================================\n');
% fprintf('Total False Positives: %4d on %4d\n',totalFalsePositives,length(CLOTHING));
fprintf('Total Model ID Classification Error : %4d on %4d | %.3f\n',totalError,ClthErrorInc,100*totalError/ClthErrorInc);
% fprintf('Total Model All Classification Error: %4d on %4d\n',allError,length(CLOTHING));
fprintf('Total Error with Clothing only: %4d on %4d | %.3f\n',totalClthError,ClthErrorInc,100*totalClthError/ClthErrorInc);
